/*
 *  mkplnrc, version 1.02 (GPL FREEWARE)
 *  Copyleft (l) Stanislav Sokolov, April 2002 and onwards.
 *
 *  This program is subject to GNU General Public License ver. 2 of June 1991
 *  and any later version. See COPYING for details.
 *
 *  You may use this source with your programs, provided
 *  due credits are given.
 *
 *  Contact the author by e-mail: stanislavs@hotmail.com
 *
 *  Internet:  http://members.tripod.com/~stanislavs/prog/prog.htm
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef NOHTTP

#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>

#endif

#include "planner.h"

#define ARGS 5

#define FOR_TAG ">Forelesninger<"
#define GRP_TAG "<h3>Gruppe"
#define TPL_TAG ">Timeplan<"

#define TEMP_FILE "mkplan.tmp"

/* Function prototypes */
char *readHTTP(const char *URL, int *size, FILE *fout);
int readln(int fd, char *buf);
char *getInfo(const char *URL);
int copyToTemp(const char *file);

int main(int argc, char **argv){
  char *buf[MAX_CR] = {NULL};
  char URL[256];
  int i, numcr = 0;
  FILE *fin = NULL, *fout = NULL;

  if(argc < ARGS){
	printf("Usage: %s -f URLfile -o outfile\n\n", argv[0]);
	return 1;
  }

  for(i = 1; i < ARGS; i+=2){
	if(argv[i][1] == 'f')
	  fin = fopen(argv[i + 1], "r");

	if(argv[i][1] == 'o')
	  fout = fopen(argv[i + 1], "w");
  }    
  
  if((fin == NULL) || fout == NULL){
	fprintf(stderr, "Can't open required files: both -f and -o must be used!\n\n");
	return 2;
  }

  /* Read the URLs and generate planner-formatted descriptions.
   * If the line does not start with http:// than it will be treated
   * as an absolute local path reference by getInfo().
   */
  
  while((fgets(URL, 255, fin) != NULL) && (numcr < MAX_CR)){
	if(URL[strlen(URL) - 1] == '\n')
	  URL[strlen(URL) - 1] = '\0';
	printf("\nProcessing: %s\n", URL);
	if((buf[numcr] = getInfo(URL)) != NULL)
	  numcr++;
  }

  fprintf(fout, "%d\n\n", numcr);

  /* Merge the descriptions into one file */
  for(i = 0; i < MAX_CR; i++){
	if(buf[i] != NULL){
	  fprintf(fout, "%s\n", buf[i]);
	  free(buf[i]);
	}
  }
  
  fclose(fin);
  fclose(fout);
  unlink(TEMP_FILE);

  return 0;
}


/*
 * This function parses a HTML page and returns a newly-allocated buffer with 
 * planner-encoded information for this course. The caller can then accumulate 
 * several such sections into one file. The caller is responsible for freeing
 * the buffer.
 */
char *getInfo(const char *URL){
  int size = 0, i, j, start, stop;
  FILE *f;
  char data[2048], line[256], tmpln[64], crname[MAX_ID - 5];
  char *retval, *tmp;

  int lectcount = 0, grpcount = 0, actcount = 0;

  /* Weekday dictionary */
  char *week[] = {"Mandag", "Tirsdag", "Onsdag", "Torsdag", "Fredag"};

  data[0] ='\0';

  /* Read the page */
  if((f = fopen(TEMP_FILE, "w")) == NULL){
	fprintf(stderr, "Error opening temporary file.\n");
	return NULL;
  }

  /* Select what to do with the URL */
  if(strncmp(URL, "http", 4) == 0){
	readHTTP(URL, &size, f);
  }else{
	size = copyToTemp(URL);
  }
  if(size == 0){
	fprintf(stderr, "No data read from network or file.\n");
	return NULL;
  }

  fclose(f);
  if((f = fopen(TEMP_FILE, "r")) == NULL){
	fprintf(stderr, "Error opening temporary file.\n");
	return NULL;
  }

  /* Generate course identifier */
  tmp = strrchr(URL, '/') + 1;
  
  i = 0;
  while(tmp[i] != '.' && tmp[i] != '\0' && i < MAX_ID - 5){
	crname[i] = toupper(tmp[i]) ;
	i++;
  }
  crname[i] = '\0';


  /* Extract information */
  while(fgets(line, 255, f) != NULL){
	if(strstr(line, FOR_TAG))
	  /* We are in lecture section */
	  break;
  }

  /* Count lectues */
  while(fgets(line, 255, f) != NULL){
	if((strncmp(line, GRP_TAG, 10) == 0) || (strstr(line, TPL_TAG) != NULL))
	  /* Done with lectures */
	  break;
	
	/* Check if the line contains information */
	for(i = 0; i < 5; i++){
	  if(strncmp(line, week[i], strlen(week[i])) == 0){ /* Meaningful line*/
	/* Read times */
	j = 0;
	while(!isdigit(line[j]) && line[j++] != '\0');
		  
	start = atoi(&line[j]);
	stop = atoi(strchr(&line[j], '-') + 1);

	sprintf(tmpln, "%d %d %d\n", i + 1, start, stop);
	strcat(data, tmpln);
	lectcount++;
	
	break;
	  } /* end if */
	} /* end for */
  }/* end while */
  

  /* Count groups */
  do{
	if(strstr(line, TPL_TAG) != NULL)
	  /* Done with groups */
	  break;

	if(strncmp(line, GRP_TAG, 10) == 0){ /* Found a group */
	  /* Read number */
	  i = atoi(strchr(line, ' '));
	  sprintf(tmpln, "\n%d\n", i);
	  strcat(data, tmpln);
	  grpcount++;
	  
	  /* Read times for activities on the next lines until blank */
	   while(fgets(line, 255, f) != NULL){
	 if(line[0] == '\n') break;
	 
	 for(i = 0; i < 5; i++){
	   if(strncmp(line, week[i], strlen(week[i])) == 0){ /* Meaningful line*/
		 /* Read times */
		 j = 0;
		 while(!isdigit(line[j]) && line[j++] != '\0');
		 
		 start = atoi(&line[j]);
		 stop = atoi(strchr(&line[j], '-') + 1);
		 
		 sprintf(tmpln, "%d %d %d\n", i + 1, start, stop);
		 strcat(data, tmpln);
		 
		 actcount++;
		 break;
	   } /* end if */
	 } /* end for */
	   } /* end activities while */
	} /* end if */       
  }while(fgets(line, 255, f) != NULL); /* end do */
  
  fclose(f);
  unlink(TEMP_FILE);

  /* Prepend header */
  sprintf(tmpln, "%s %d %d %d\n", crname, lectcount, grpcount, (grpcount == 0?0:actcount/grpcount));
  retval = malloc(strlen(data) + strlen(tmpln) + 1);
  strcpy(retval, tmpln);
  strcat(retval, data);

  return retval;
}

#ifdef NOHTTP

char *readHTTP(const char *URL, int *size, FILE *fout){
  fprintf(stderr, "HTTP support is not compiled!\n");
  *size = 0;
  return NULL;
}

#else

/*
 * This function is a client-side implementation of HTTP 1.0 protocol that
 * reads a well-formed HTTP URL.
 * If fout if not NULL, the data will be written to that file, otherwise
 * a new buffer is alocated and filled with data retrieved. It is caller
 * responsibility to free that buffer.
 * size will contain the number of bytes read from server.
 */
char *readHTTP(const char *URL, int *size, FILE *fout){
  char *buf = NULL;
  char server[128] = {'\0'};
  char head[512] = {'\0'};
  char req[128];
  
  struct in_addr hostaddr;
  struct sockaddr_in address;
  struct hostent *hinfo;
  int sock, i, bytes, tmp;

  *size = 0;

  /* Check URL protocol */
  if(strncmp(URL, "http://", 7) != 0){
	return NULL;
  }

  fprintf(stderr, "Connecting to host... ");
  /* Get server name or address */
  strncpy(server, URL + 7, (strchr(URL + 7, '/') - (URL + 7)));

  hinfo = gethostbyname(server);
  if(!hinfo){
	perror("gethostbyname");
	return NULL;
  }
  memcpy(&hostaddr, *(hinfo->h_addr_list), sizeof(struct in_addr));

  /* Open connection */
  memset(&address, 0, sizeof address);
  address.sin_family = AF_INET;
  address.sin_addr = hostaddr;
  address.sin_port = htons(80);
  
  if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
	perror("socket");
	return NULL;
  }
  
  if(connect(sock,(struct sockaddr*) &address, sizeof address) < 0){
	perror("connect");
	return NULL;
  }
  fprintf(stderr, "done.\n");

  /* Send request */
  fprintf(stderr, "Sending request... ");
  sprintf(req, "GET %s HTTP/1.0\n"
	  "User-Agent: Planner/1.01\n"
	  "\n", strchr(URL + 7, '/'));
  if(write(sock, req, strlen(req) + 1) < 0){
	perror("request");
	return NULL;
  }
  fprintf(stderr, "done.\n");

  /* Get response */
  /* Read header and make some checks */
  fprintf(stderr, "Recieving header... ");
  bytes = 0;
  do{
	if(readln(sock, head) == -1){
	  perror("read header");
	  return NULL;
	}

	if(strstr(head, "HTTP") != NULL){
	  if(strstr(head, "200 OK") == NULL){
	fprintf(stderr, "%s", head);
	return NULL;
	  }
	}

	if(strstr(head, "Content-Length") != NULL){
	  bytes = atoi(strchr(head, ':') + 1);
	}
  }while(head[0] != '\r');
  
  *size = bytes;
  if(bytes == 0){
	fprintf(stderr, "Invalid response from server.\n");
	return NULL;
  }
  fprintf(stderr, "%d bytes to be read.\n", bytes);

  /* Read data */
  fprintf(stderr, "Reading data... ");

  /* read data to buffer in maximum chunks */
  buf = malloc(bytes + 1);
  i = 0;
  while((tmp = read(sock, buf + i, bytes)) > 0){ 
	i += tmp;
  }
  
  if(fout != NULL){
	fprintf(fout, "%s", buf);
	free(buf);
	buf = NULL;
  }

  fprintf(stderr, "done.\n");

  /* The socket is closed by server */

  return buf;

}

#endif


/*
 * A helper-function used by readHTTP that gives a gets() functionality on
 * file descriptors.
 */
int readln(int fd, char *buf){
  int i = 0;
  int ret;
  do{
	ret = read(fd, &buf[i], 1);
  }while(buf[i++] != '\n');
  buf[i] = '\0';

  return ret;  
}


int copyToTemp(const char *file){
  FILE *fin, *fout;
  char c;
  int counter = 0;

  if((fin = fopen(file, "r")) == NULL)
	return 0;

  if((fout = fopen(TEMP_FILE, "w")) == NULL)
	return 0;

  while((c = fgetc(fin)) != EOF){
	counter++;
	fprintf(fout, "%c", c);
  }

  fclose(fin);
  fclose(fout);

  fprintf(stderr, "Copied %d bytes.\n", counter);

  return counter;
}


