/*
 *  Planner, version 1.02 (GPL FREEWARE)
 *  Copyleft (l) Stanislav Sokolov, June 2000 and onwards.
 *
 *  This program is subject to GNU General Public License ver. 2 of June 1991
 *  and any later version. See COPYING for details.
 *
 *  You may use this source with your programs, provided
 *  due credits are given.
 *
 *  Contact the author by e-mail: stanislavs@hotmail.com
 *
 *  Internet:  http://members.tripod.com/~stanislavs/prog/prog.htm
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "planner.h"

//Use -DPLAIN switch when compiling on UNIX systems, and systems that
//do no support Extended ASCII pseudographics.
//#define PLAIN

#ifdef PLAIN

#define TOP "+--+--------------+--------------+--------------+--------------+--------------+\n"
#define WK  "|Hr|   Monday     |   Tuesday    |  Wednesday   |   Thursday   |    Friday    |\n"
#define MID "+--+--------------+--------------+--------------+--------------+--------------+\n"
#define TXT "|%02d|%14s|%14s|%14s|%14s|%14s|\n"
#define BOT "+--+--------------+--------------+--------------+--------------+--------------+\n\n"

#else

#define TOP "Ŀ\n"
#define WK  "Hr   Monday        Tuesday      Wednesday      Thursday       Friday    \n"
#define MID "Ĵ\n"
#define TXT "%02d%14s%14s%14s%14s%14s\n"
#define BOT "\n\n"

#endif

#define STAT "\n%lu tables: <= %d overlap; %d-%d (+-%d hours); min. %d freedays; max. %d \"holes\".\n\n"


//Flags passed to the rewind function of Course class.
#define REW_L     0
#define REW_G     1


struct Filter{
  int maxOverlap;
  int iStart;
  int iStop;
  int dev;
  int minFree;
  int holes;
};


class Group{
protected:
  int cur;
  int weekday[MAX_ACT];
  int on[MAX_ACT];
  int off[MAX_ACT];
  char id[MAX_ID];
  
public:
  Group(void);
  int add(int w, int start, int stop);
  int getWeek(int num);
  int getStart(int num);
  int getStop(int num);
  void putId(char desc[]);
  char *getId(void);
};

class Lect{
protected:
  int weekday;
  int on;
  int off;
  
public:
  Lect(void);
  void add(int w, int start, int stop);
  int getWeek(void);
  int getStart(void);
  int getStop(void);
};

class Course{
protected:
  Group grp[MAX_GRP];
  Lect  lect[MAX_LECT];
  int numLect, curLect;
  int numGrp, curGrp;
  char id[10];

public:
  Course(void);
  int add(int w, int start, int stop);
  int add(int num, int w[], int start[], int stop[]);
  void rewind(int what);
  Group *getNextGrp(void);
  Lect  *getNextLect(void);
  void  putId(char name[]);
  int getTotGrp(void);
  int getCurGrp(void);
  int getTotLect(void);
  char  *getId(void);
};


//Function prototypes
char Ask(const char *Prompt);
void lnerr(FILE *file, int linenum);
unsigned long  readLn(char *buf, int len, FILE *file);
void reg(Course *cr, FILE *file);
Filter getFilter(void);
int doWrite(char table[5][12][15], Filter filt);
int makeTable(Course *cr, char table[5][12][15], Group *grp[], Filter filt);
void analyse(Course *cr, FILE *file, Filter filt, char fStdin);


//Group member functions
Group::Group(void){
  for(int i = 0; i < MAX_ACT; i++){
    weekday[i] = 0;
    on[i] = 0;
    off[i] = 0;
  }

  cur = 0;
  id[0] = '\0';
}

int Group::add(int w, int start, int stop){
  if(cur == MAX_ACT) return 1;
  
  weekday[cur] = w;
  on[cur] = start;
  off[cur] = stop;
  
  cur++;
  
  return 0;
}

int Group::getWeek(int num){
  return weekday[num];
}

int Group::getStart(int num){
  return on[num];
}

int Group::getStop(int num){
  return off[num];
}

char *Group::getId(void){
  return id;
}

void Group::putId(char desc[]){
  strcpy(id, desc);
}


//Lect member functions
Lect::Lect(void){
  weekday = 0;
  on = 0;
  off = 0;
}

void Lect::add(int w, int start, int stop){
  weekday = w;
  on = start;
  off = stop;
}

int Lect::getWeek(void){
  return weekday;
}

int Lect::getStart(void){
  return on;
}

int Lect::getStop(void){
  return off;
}


//Course member functions
Course::Course(void){
  numLect = curLect = 0;
  numGrp = curGrp = 0;
  id[0] = '\0';
}


int Course::add(int w, int start, int stop){
  if(numLect == MAX_LECT) return 1;
  
  lect[curLect].add(w, start, stop);
  curLect++;
  numLect++;

  return 0;
}

int Course::add(int num, int w[], int start[], int stop[]){
  if(numGrp == MAX_GRP) return 1;
  
  char desc[MAX_ID];
  
  sprintf(desc, "%s-%d", id, num);
  
  grp[curGrp].putId(desc);
  for(int i = 0; i < MAX_ACT; i++)
    grp[curGrp].add(w[i], start[i], stop[i]);
  
  curGrp++;
  numGrp++;
  
  return 0;
}

void Course::rewind(int what){
  if(what == REW_L)
    curLect = 0;
  else if(what == REW_G)
    curGrp = 0;
}


Group *Course::getNextGrp(void){
  return &grp[curGrp++];
}

Lect  *Course::getNextLect(void){
  return &lect[curLect++];
}

void  Course::putId(char name[]){
  strcpy(id, name);
}

int Course::getTotGrp(void){
  return numGrp;
}

int Course::getCurGrp(void){
  return curGrp;
}

int Course::getTotLect(void){
  return numLect;
}

char  *Course::getId(void){
  return id;
}



//General prompt routine
char Ask(const char *Prompt){
  char Resp, r[3];
  
  fflush(stdin);

  fprintf(stderr, "%s ", Prompt);
  
  fgets(r, 3, stdin);
  Resp = r[0];

  if((Resp >= 97) && (Resp <= 122)) Resp -= 32;
  
  if(Resp == 'Q') exit(1);  //User wants to quit
  
  return Resp;
}



//This one is called when something is wrong in reg()
void lnerr(FILE* file, int linenum){
  fprintf(stderr, "Error in line %d.\n\n", linenum);
  if(file != stdin) fclose(file);
  exit(2);
}



//Reads a line from the named file and returns the number of the currently
//returned line.
unsigned long readLn(char *buf, int len, FILE *file){
  static unsigned long linenum = 0;
  char *x;

  fflush(file);
  do{
    x = fgets(buf, len, file);
    linenum++;
  }while((x != NULL) && ((buf[0] == '#') || (buf[0] == '\n')));

  return linenum;
}



//Registers data about the courses structure, given in an input file.
void reg(Course *cr, FILE *file){
  char id[10];
  char line[256];
  unsigned long linenum;
  int num, w[MAX_ACT], start[MAX_ACT], stop[MAX_ACT];
  int cMax = 0, lMax = 0, gMax = 0, aMax = 0, i;

  linenum = readLn(line, 256, file);
  sscanf(line, "%d", &cMax);
  if((cMax <= 0) || (cMax > MAX_CR)) lnerr(file, linenum);

  for(int curCr = 0; curCr < cMax; curCr++){
    //Reset activity tables
    for(i = 0; i < MAX_ACT; i++){
      w[i] = 0;
      start[i] = 0;
      stop[i] = 0;
    }

    //'Course's name' 'numlect' 'numgrp' 'numact'
    linenum = readLn(line, 256, file);
    sscanf(line, "%9s %d %d %d", id, &lMax, &gMax, &aMax);
    cr[curCr].putId(id);
    if((lMax < 0) || (lMax > MAX_LECT)) lnerr(file, linenum);
    if((gMax < 0) || (gMax > MAX_GRP)) lnerr(file, linenum);
    if((aMax < 0) || (aMax > MAX_ACT)) lnerr(file, linenum);

    //Read lectures
    if(lMax > 0){
      for(i = 0; i < lMax; i++){
	linenum = readLn(line, 256, file);
	sscanf(line, "%d %d %d", &w[0], &start[0], &stop[0]);
	if((w[0] < 1) || (w[0] > 5)) lnerr(file, linenum);
	if((start[0] < 8) || (start[0] > 19)) lnerr(file, linenum);
	if((stop[0] < 9) || (stop[0] > 20)) lnerr(file, linenum);

	cr[curCr].add(w[0] - 1, start[0], stop[0]);
      }
    }

    //Read groups
    if(gMax > 0){
      for(int j = 0; j < gMax; j++){
				//Group header
	linenum = readLn(line, 256, file);
	sscanf(line, "%d", &num);
	if((num < 1) || (num > 999)) lnerr(file, linenum);

				//Read activities in the group
	for(int k = 0; k < aMax; k++){
	  linenum = readLn(line, 256, file);
	  sscanf(line, "%d %d %d", &w[k], &start[k], &stop[k]);
	  if((w[k] < 1) || (w[k] > 5)) lnerr(file, linenum);
	  w[k]--;
	  if((start[k] < 8) || (start[k] > 20)) lnerr(file, linenum);
	  if((stop[k] < 8) || (stop[k] > 20)) lnerr(file, linenum);
	}

	cr[curCr].add(num, w, start, stop);
      }
    }
  }
}



//Gets user parameters as to how the program should filter the data
Filter getFilter(void){
  Filter filt;
  char buf[4];

  do{
    filt.maxOverlap = Ask("\nGive the maximum number of overlaps (0-9)") - 48;
  }while((filt.maxOverlap < 0) || (filt.maxOverlap > 9));

  fprintf(stderr, "\nGive the interval for the activities.\n"
	  "(Lectures are permanent and are not counted at interval filtering.)\n");
  do{
    fflush(stdin);
    fprintf(stderr, "\nStart (8-19): ");
    fgets(buf, 4, stdin);
    filt.iStart = atoi(buf);
  }while((filt.iStart < 8) || (filt.iStart > 19));

  do{
    fflush(stdin);
    fprintf(stderr, "End (9-20): ");
    fgets(buf, 4, stdin);
    filt.iStop = atoi(buf);
  }while((filt.iStop < 9) || (filt.iStop > 20));

  do{
    filt.dev = Ask("\nHow large deviation from this interval (in hours/week) can you accept? (0-9)") - 48;
  }while((filt.dev < 0) || (filt.dev > 9));

  do{
    filt.minFree = Ask("\nHow many free days would you like to have? (0-4)") - 48;
  }while((filt.minFree < 0) || (filt.minFree > 4));

  do{
    filt.holes = Ask("\nHow many \"hole\"-hours can you accept between activities? (0-9)") - 48;
  }while((filt.holes < 0) || (filt.holes > 9));

  return filt;
}



//See if the table shall be outputed, given the current filter.
int doWrite(char table[5][12][15], Filter filt){
  char i, j, w, start, stop, weights[12], weight;

  //Check distribution
  weight = 0;
  for(i = 0; i < 12; i++){ //For each hour...
    weights[i] = 0;
    for(j = 0; j < 5; j++) //...in all weekdays...
      if((table[j][i][0] != '\0') && (table[j][i][0] != 'l') && (table[j][i][0] != '*'))
	weights[i]++;      //...calculate the total number of group activities.
  }


  //See how many total hours go before the user's preferenced start.
  for(i = 0; i < filt.iStart - 8; i++)
    weight += weights[i];

  //See how many total hours go before the user's preferenced end.
  for(i = filt.iStop - 8; i < 12; i++)
    weight += weights[i];

  //See if this deviation is greater than the user's preference.
  if(weight > filt.dev)
    return FALSE;


  //Check for free days
  weight = 0;
  for(i = 0; i < 5; i++){
    w = 0;
    for(j = 0; j < 12; j++)
      if(table[i][j][0] != '\0')
	w++;

    if(w == 0) weight++;
  }

  if(weight < filt.minFree)
    return FALSE;


  //Check for "holes"
  for(j = 0; j < 5; j++){
    weight = 0;
    start = 1; //Must be initially 1 to avoid confusing (start == stop == 0)
    stop = 0;  //cases of one activity at 8:00 and no activities in the day.

    //Find where activities start
    for(i = 0; i < 12; i++){
      if(table[j][i][0] != '\0'){
	start = i;
	break;
      }
    }

    //Find where activities end
    for(i = 11; i >= 0; i--){
      if(table[j][i][0] != '\0'){
	stop = i;
	break;
      }
    }

    //See if there are any blanks between start and end.
    for(i = start; i <= stop; i++)
      if(table[j][i][0] == '\0') weight++;

    if(weight > filt.holes)
      return FALSE;
  }

  return TRUE;
}


//Table generation function, which uses cut-off if the primary condition
//of the amount of overlaps is not met.
int makeTable(Course *cr, char table[5][12][15], Group *grp[], Filter filt){
  int overlap = 0;           //Primary condition
  int start, stop, w;        //Temporary variables
  register char i, j, k, l;  //Counters
  Lect *lect;

  //Fill lectures.
  for(i = 0; i < MAX_CR; i++){ //Courses
    cr[i].rewind(REW_L);
    for(j = 0; j < cr[i].getTotLect(); j++){ //Lectures
      lect = cr[i].getNextLect();
      w = lect->getWeek();
      start = lect->getStart();
      stop = lect->getStop();

      //Go through the lecture intervall and see if it
      //collides with something.
      for(k = start - 8; k < stop - 8; k++){
	if(table[w][k][0] == 'l')
	  strcpy(table[w][k], "*");
	else
	  strcpy(table[w][k], "l");

	strcat(table[w][k], cr[i].getId());
      }
    }
  }

  //Fill groups.
  for(i = 0; i < MAX_CR; i++){ //Courses
    if(cr[i].getTotGrp() > 0){ //Is there something to do for this course?
      for(l = 0; l < MAX_ACT; l++){ //Activities in the group
	w = grp[i]->getWeek(l);
	start = grp[i]->getStart(l);
	if(start == 0) break;
	stop = grp[i]->getStop(l);

				//Go through the activity intervall and see if it
				//collides with something.
	for(k = start - 8; k < stop - 8; k++){
	  if(table[w][k][0] != '\0'){
	    if((table[w][k][0] == 'g') || (table[w][k][0] == '+')){
	      //If it is another group, mark it and overwrite
	      strcpy(table[w][k], "+");
	      strcat(table[w][k], grp[i]->getId());
	      overlap++;
	      if(overlap > filt.maxOverlap) return overlap;
	    }else if((table[w][k][0] == 'l') || (table[w][k][0] == '*') || (table[w][k][0] == '!')){
	      //If it is a lecture, mark it and do nothing else
	      table[w][k][0] = '!';
	      overlap++;
	      if(overlap > filt.maxOverlap) return overlap;
	    }
	  }else{
	    //No collision? Good!
	    strcpy(table[w][k], "g");
	    strcat(table[w][k], grp[i]->getId());
	  }
	}// end for(k = start.....
      }//end for(l = 0......
    }//end if(cr[i]........
  }//end for(i = 0........

  return overlap;
}



//Main workhorse
void analyse(Course *cr, FILE *file, Filter filt, char fStdin){
  unsigned long comb = 1, m, count = 0;
  register char i, j;        //Counters
  char carry = 0;            //Used to calculate the next group combination.
  char table[5][12][15];     //Weekdays, hours, text
  Group *grp[MAX_CR]; //An active group from each Course

  //For all Courses...
  for(i = 0; i < MAX_CR; i++){
    //Initialize groups
    if(cr[i].getId()[0] == '\0') break; //No more Courses registered.

    cr[i].rewind(REW_G);
    if(cr[i].getTotGrp() > 0){
      grp[i] = cr[i].getNextGrp();
      if(cr[i].getCurGrp() == cr[i].getTotGrp())
	cr[i].rewind(REW_G);

      //Count the number of combnations (a product of all groups).
      //This is the number of all possible tables.
      comb *= cr[i].getTotGrp();
    }
  }


  for(m = 0; m < comb; m++){
    //Reset the table
    for(i = 0; i < 5; i++)
      for(j = 0; j < 12; j++)
	table[i][j][0] = '\0';

    //Generate the table, check the filters and make an output.
    if(makeTable(cr, table, grp, filt) <= filt.maxOverlap){
      if(doWrite(table, filt)){
	count++; //Count outputted tables.

				//Header
	fprintf(file, "The first symbol has a special meaning:\n"
		"g - group; l - lecture; * - overlap between lectures\n"
		"+ - overlap between groups; ! - overlap between a lecture and a group\n\n");
	fprintf(file, TOP);
	fprintf(file, WK);
	fprintf(file, MID);

				//Body
	for(i = 0; i < 12; i++)
	  fprintf(file, TXT, i + 8, table[0][i], table[1][i], table[2][i], table[3][i], table[4][i]);

				//Trailer
	fprintf(file, BOT);
      }
    }

    //Set up group data for the next table (iteration)
    carry = 1; //Used to rotate the groups
    for(i = 0; i < MAX_CR; i++){
      if(carry){
	carry = 0;

	if(cr[i].getTotGrp() > 0)
	  grp[i] = cr[i].getNextGrp();

				//If it is the last group in the current Course's list, rewind
				//to the beginning of the list so as to start over from the next
				//iteration.
	if(cr[i].getCurGrp() == cr[i].getTotGrp()){
	  cr[i].rewind(REW_G);
	  carry = 1;
	}
      }
    }//end for(i = 0,......

    if(!fStdin && (count != 0) && (count % MAX_TBLS == 0)){
      fprintf(stderr, "%lu tables that meet you filter specifications are generated.\n"
	      "Continue? (Y/N)", count);
      if(Ask("") == 'N') break;
    }

  }//end for(m = 0; m < comb.....

  fprintf(stderr, STAT, count, filt.maxOverlap, filt.iStart, filt.iStop, filt.dev, filt.minFree, filt.holes);
  if(count != 0)	fprintf(file, STAT, count, filt.maxOverlap, filt.iStart, filt.iStop, filt.dev, filt.minFree, filt.holes);

}



int main(int argc, char *argv[]){
  char repl, fStdin;
  FILE *f;
  unsigned long linenum;
  char line[256];
  Filter filt;
  Course cr[MAX_CR];

  f = stdin; //Default
  if(argc > 1){
    if(!(strcmp(argv[1], "--help") && strcmp(argv[1], "/?") && strcmp(argv[1], "-?"))){
      fprintf(stderr, "\n\nPlanner, version 1.01 (GPL FREEWARE)\n"
	      "Copyleft (l) Stanislav Sokolov, June 2000 and onwards.\n\n"

	      "The program will generate a set of optimal timetables, based on the information\n"
	      "about the study courses and a set of filters. The system is primarily designed\n"
	      "to comply with course layout at the University in Oslo, but can be tailored to\n"
	      "other institutions, which have lectures and a number group activities,\n"
	      "assosiated with each course.\n\n"

	      "The information about courses is given via a script-file. The layout of this\n"
	      "file is found in SAMPLE.TXT. The script can be given as the first arguement or\n"
	      "redirected to stdin (in which case the filsers must also be present in the\n"
	      "script), if the first arguement is omitted or the filename is invalid. The\n"
	      "output is written to the file, named by the second arguement or to stdout if no\n"
	      "or invalid output filename is given.\n\n"
	      "Sample usage:\n"
	      "  planner sample.txt out.txt       (filters will be prompted)\n"
	      "  planner < sample.txt > out.txt   (filters from sample.txt will be used)\n"
	      "  planner . out.txt < sample.txt   (the same as above, only output is written\n"
	      "                                   to file and not stdout)\n\n");
      return 0;
    }

    //Find where to read data from
    if((f = fopen(argv[1], "r")) == NULL) //Try the first arguement
      f = stdin;

  }

  if(f == stdin)
    fStdin = TRUE;
  else
    fStdin = FALSE;



  //Read data
  fprintf(stderr, "Registering courses...\n");
  reg(cr, f);
  fprintf(stderr, "Finished registering courses.\n");

  if(f != stdin) fclose(f);

  //Find where to write output to
  f = stdout;
  if(argc > 2){
    if((f = fopen(argv[2], "a+t")) == NULL)
      f = stdout;
  }

  //Get and apply filters
  do{
    if(fStdin){ //In most cases this would mean the input was redirected.
      fprintf(stderr, "\nReading filters...\n");
      linenum = readLn(line, 256, stdin);
      //Read the filters from line
      sscanf(line, "%d %d %d %d %d %d %c", &filt.maxOverlap,
	     &filt.iStart,
	     &filt.iStop,
	     &filt.dev,
	     &filt.minFree,
	     &filt.holes,
	     &repl);
      //Check filter validity
      if((filt.maxOverlap < 0) || (filt.maxOverlap > 9) ||
         (filt.iStart < 8) || (filt.iStart > 19) ||
         (filt.iStop < 9) || (filt.iStop > 20) ||
         (filt.dev < 0) || (filt.dev > 9) ||
         (filt.minFree < 0) || (filt.minFree > 4) ||
         (filt.holes < 0) || (filt.holes > 9)
	 ) lnerr(stdin, linenum);

    }else
      filt = getFilter();


    fprintf(stderr, "\n\nWorking . . .\n");

    analyse(cr, f, filt, fStdin);

    //Get the continuation condition
    if(fStdin){
      if((repl >= 97) && (repl <= 122)) repl -= 32;
      if((repl != 'Y') && (repl != 'N')) repl = 'N';
    }else{
      fprintf(stderr, "\n\n\nYou can generate more tabels using the same data,\n"
	      "while applying a new set of filters.\n");
      do{
	repl = Ask("\nDo you want to do it? (Y/N)");
      }while((repl != 'Y') && (repl != 'N'));
    }

  }while(repl == 'Y');

  if(f != stdout) fclose(f);

  return 0;
}
